@@ -44,7 +44,7 @@ def pytest_configure(config):
     )
     config.addinivalue_line("markers", config_line)
 
-    if config.getoption("indulgent-ordering"):
+    if config.getoption("indulgent_ordering"):
         # We need to dynamically add this `tryfirst` decorator to the plugin:
         # only when the CLI option is present should the decorator be added.
         # Thus, we manually run the decorator on the class function and
@@ -65,14 +65,22 @@ def pytest_addoption(parser):
     """Set up CLI option for pytest"""
     group = parser.getgroup("ordering")
     group.addoption("--indulgent-ordering", action="store_true",
-                    dest="indulgent-ordering",
+                    dest="indulgent_ordering",
                     help="Request that the sort order provided by "
                          "pytest-order be applied before other sorting, "
                          "allowing the other sorting to have priority")
     group.addoption("--order-scope", action="store",
-                    dest="order-scope",
-                    help="Defines the scope used for ordeing. Possible values"
+                    dest="order_scope",
+                    help="Defines the scope used for ordering. Possible values"
                          "are 'session' (default), 'module', and 'class'")
+    group.addoption("--sparse-ordering", action="store_true",
+                    dest="sparse_ordering",
+                    help="If there are gaps between ordinals they are filled "
+                         "with unordered tests.")
+    group.addoption("--order-dependencies", action="store_true",
+                    dest="order_dependencies",
+                    help="If set, dependencies added by pytest-dependency will"
+                         "be ordered if needed.")
 
 
 class OrderingPlugin(object):
@@ -84,6 +92,26 @@ class OrderingPlugin(object):
     """
 
 
+class Settings:
+    sparse_ordering = False
+    order_dependencies = False
+    scope = "session"
+
+    @classmethod
+    def initialize(cls, config):
+        cls.sparse_ordering = config.getoption("sparse_ordering")
+        cls.order_dependencies = config.getoption("order_dependencies")
+        scope = config.getoption("order_scope")
+        if scope in ("session", "module", "class"):
+            cls.scope = scope
+        else:
+            if scope is not None:
+                warn("Unknown order scope '{}', ignoring it. "
+                     "Valid scopes are 'session', 'module' and 'class'."
+                     .format(scope))
+            cls.scope = "session"
+
+
 def get_filename(item):
     name = item.location[0]
     if os.sep in name:
@@ -91,10 +119,28 @@ def get_filename(item):
     return name[:-3]
 
 
-def mark_binning(item, keys, start, end, before, after, unordered):
+def mark_binning(item, keys, start, end, before, after, unordered, alias):
+    if ("dependency" in keys and
+            (Settings.order_dependencies or "order" in keys)):
+        # always order dependencies if an order mark is present
+        # otherwise only if order-dependencies is set
+        mark = item.get_closest_marker("dependency")
+        prefix = get_filename(item) + "."
+        dependent_mark = mark.kwargs.get("depends")
+        if dependent_mark:
+            for name in dependent_mark:
+                if "." not in name:
+                    name = prefix + name
+                after.setdefault(name, []).append(item)
+        name_mark = mark.kwargs.get("name")
+        if name_mark:
+            alias[prefix + name_mark] = prefix + item.name
+
     if "order" in keys:
         mark = item.get_closest_marker("order")
         order = mark.args[0] if mark.args else None
+        if order is None:
+            order = mark.kwargs.get("index")
         before_mark = mark.kwargs.get("before")
         after_mark = mark.kwargs.get("after")
         if order is not None:
@@ -126,24 +172,23 @@ def mark_binning(item, keys, start, end, before, after, unordered):
     return False
 
 
-def insert(items, sort):
-    if isinstance(items, tuple):
-        list_items = items[1]
-    else:
-        list_items = items
-    sort += list_items
-
-
 def insert_before(name, items, sort):
     regex_name = re.escape(name) + r"(:?\.\w+)?$"
     for pos, item in enumerate(sort):
         prefix = get_filename(item)
         item_name = prefix + "." + item.location[2]
         if re.match(regex_name, item_name):
-            if pos == 0:
-                sort[:] = items + sort
-            else:
-                sort[pos:1] = items
+            for item_to_insert in items:
+                if item_to_insert in sort:
+                    index = sort.index(item_to_insert)
+                    if index > pos:
+                        del sort[index]
+                        sort.insert(pos, item_to_insert)
+                else:
+                    if pos == 0:
+                        sort[:] = items + sort
+                    else:
+                        sort[pos:1] = items
             return True
     return False
 
@@ -151,12 +196,18 @@ def insert_before(name, items, sort):
 def insert_after(name, items, sort):
     regex_name = re.escape(name) + r"(:?\.\w+)?$"
     for pos, item in reversed(list(enumerate(sort))):
-        prefix = get_filename(item)
-        item_name = prefix + "." + item.location[2]
+        item_name = get_filename(item) + "." + item.location[2]
         if re.match(regex_name, item_name):
-            sort[pos + 1:1] = items
+            for item_to_insert in items:
+                if item_to_insert in sort:
+                    index = sort.index(item_to_insert)
+                    if index < pos + 1:
+                        del sort[index]
+                        pos -= 1
+                        sort.insert(pos + 1, item_to_insert)
+                else:
+                    sort[pos + 1:1] = items
             return True
-
     return False
 
 
@@ -166,21 +217,16 @@ def do_modify_items(items):
     start_item = {}
     end_item = {}
     unordered_list = []
+    alias_names = {}
 
     for item in items:
         mark_binning(item, item.keywords.keys(), start_item, end_item,
-                     before_item, after_item, unordered_list)
+                     before_item, after_item, unordered_list, alias_names)
 
     start_item = sorted(start_item.items())
     end_item = sorted(end_item.items())
 
-    sorted_list = []
-
-    for entries in start_item:
-        insert(entries, sorted_list)
-    insert(unordered_list, sorted_list)
-    for entries in end_item:
-        insert(entries, sorted_list)
+    sorted_list = sort_numbered_items(start_item, end_item, unordered_list)
 
     still_left = 0
     length = len(before_item) + len(after_item)
@@ -196,7 +242,8 @@ def do_modify_items(items):
 
         remove_labels = []
         for label, after in after_item.items():
-            if insert_after(label, after, sorted_list):
+            name = alias_names[label] if label in alias_names else label
+            if insert_after(name, after, sorted_list):
                 remove_labels.append(label)
         for label in remove_labels:
             del after_item[label]
@@ -216,17 +263,34 @@ def do_modify_items(items):
     return sorted_list
 
 
+def sort_numbered_items(start_item, end_item, unordered_list):
+    sorted_list = []
+    index = 0
+    for entries in start_item:
+        if Settings.sparse_ordering:
+            while entries[0] > index and unordered_list:
+                sorted_list.append(unordered_list.pop(0))
+                index += 1
+        sorted_list += entries[1]
+        index += len(entries[1])
+    mid_index = len(sorted_list)
+    index = -1
+    for entries in reversed(end_item):
+        if Settings.sparse_ordering:
+            while entries[0] < index and unordered_list:
+                sorted_list.insert(mid_index, unordered_list.pop())
+                index -= 1
+        sorted_list[mid_index:mid_index] = entries[1]
+        index -= len(entries[1])
+    sorted_list[mid_index:mid_index] = unordered_list
+    return sorted_list
+
+
 def modify_items(session, config, items):
-    scope = config.getoption("order-scope")
-    if scope not in ("session", "module", "class"):
-        if scope is not None:
-            warn("Unknown order scope '{}', ignoring it. "
-                 "Valid scopes are 'session', 'module' and 'class'."
-                 .format(scope))
-        scope = "session"
-    if scope == "session":
+    Settings.initialize(config)
+    if Settings.scope == "session":
         sorted_list = do_modify_items(items)
-    elif scope == "module":
+    elif Settings.scope == "module":
         module_items = OrderedDict()
         for item in items:
             module_path = item.nodeid[:item.nodeid.index("::")]
